require_dependency "binect_importer/application_controller"
require_dependency "../../lib/binect_importer/googledrive_importer"

require 'google/apis/drive_v3'
require 'google/api_client/client_secrets'

# This configuration will be added by the Rails-Application
import_config = Rails.application.config.import_config[:googledrive]
ROOT_PATH     = import_config[:ROOT_PATH]

module BinectImporter
    class GoogledriveRegistrationsController < AuthenticatedController
        def start_registration
            auth_client = get_auth_client
            auth_client.update!(
                :scope => 'https://www.googleapis.com/auth/drive https://www.googleapis.com/auth/drive.metadata',
                :redirect_uri => 'https://web.binect.de/googledrive_registrations/finish_registration',
                :additional_parameters => {
                    "access_type" => "offline",         # offline access
                    "include_granted_scopes" => "true"  # incremental auth
                }
            )
            auth_uri = auth_client.authorization_uri.to_s
            Rails.logger.info "Start GoogleDrive registration for user: #{current_user}. Now redirect to URL: #{auth_uri}"

            redirect_to auth_uri
        end

        # GET /import_registration/dropbox_finish
        def finish_registration
            Rails.logger.info "Registration succeeded for user: #{current_user}."

            if params[:error].present?
                Rails.logger.info "GoogleDrive Registration failed: #{params[:error]}"
            else
                auth_code = params[:code]

                # Get access token for authentication code
                begin
                    auth_client = get_auth_client
                    auth_client.code = auth_code
                    auth_client.fetch_access_token!

                    # Check whethere this user is already authorized
                    importer_data_index = current_user.authenticated_importers.index { |importer_hash| importer_hash[:type] == GoogledriveImporter::IMPORTER_TYPE }
                    importer_data = { class_name: "BinectImporter::GoogledriveImporter", type: GoogledriveImporter::IMPORTER_TYPE, access_data: auth_client.to_json }

                    if importer_data_index.nil?
                        # Add new importer_data
                        Rails.logger.info "Saving new access token"
                        current_user.authenticated_importers.push(importer_data)
                    else
                        # Replace importer data
                        Rails.logger.info "Updating new access token"
                        current_user.authenticated_importers[importer_data_index] = importer_data
                    end
                    unless current_user.save
                        Rails.logger.error  "Unable to persist change: #{current_user.errors}"
                        flash[:error] = "Unable to save data. Please try again"
                    end
                rescue => e
                    flash[:error] = "Error: #{e}"
                end
            end

            redirect_to ROOT_PATH
        end

        def unlink
            importer_data = current_user.authenticated_importers.select { |importer_hash| importer_hash[:type] == GoogledriveImporter::IMPORTER_TYPE }
            auth_client = get_auth_client
            access_data_hash = JSON.parse(importer_data.first[:access_data])

            begin
                auth_client = Signet::OAuth2::Client.new(access_data_hash)
                auth_client.fetch_access_token!

                uri = URI('https://accounts.google.com/o/oauth2/revoke')
                params = { :token => auth_client.access_token }
                uri.query = URI.encode_www_form(params)
                Net::HTTP.get(uri)

                flash[:success] = "Unlinking successfully"
            rescue => e
                flash[:error] = "Error: #{e}"
            ensure
                # Always remove data.
                importer_data_index = current_user.authenticated_importers.index { |importer_hash| importer_hash[:type] == GoogledriveImporter::IMPORTER_TYPE }
                current_user.authenticated_importers.delete_at(importer_data_index)
                current_user.save
            end

            redirect_to ROOT_PATH
        end

        private
        def get_auth_client
            Google::APIClient::ClientSecrets.load( File.join(Rails.root,'config','googledrive_config.json')).to_authorization
        end
    end
end
