require 'google/apis/drive_v3'
require 'google/api_client/client_secrets'

# see:  https://developers.google.com/drive/v3/web/folder
#       http://www.rubydoc.info/github/google/google-api-ruby-client/Google/Apis/DriveV3/DriveService
#
# Create Folder:
# file_metadata = {name: 'Invoices',mime_type: 'application/vnd.google-apps.folder'}
# file = drive.create_file(file_metadata, fields: 'id')
# puts "Folder Id: #{file.id}"
module BinectImporter
    class GoogledriveImporter
        IMPORTER_TYPE = "googledrive"

        def initialize(user)
            @user = user
        end

        def import
            drive = drive_service
            folder_id = create_folder( 'BinectImporter', drive )

            page_token = nil
            begin
                response = drive_service.list_files(q: "mimeType='application/pdf' and '#{folder_id}' in parents",
                                                    spaces: 'drive',
                                                    fields:'nextPageToken, files(id, name)',
                                                    page_token: page_token)
                for file in response.files
                    # Process change
                    puts "Found file: #{file.name} #{file.id}"
                    Rails.logger.info "Import PDF: Load file: #{file.name} #{file.id}"

                    # Load Meta-Data
                    file = drive_service.get_file(file.id)

                    # Load File-Content
                    file_buffer = drive_service.get_file(file.id, download_dest: StringIO.new)

                    Rails.logger.info "Import PDF: ready"

                    # Call external block that will handle the file.
                    is_success = yield( file_buffer, file.name, file.mime_type, file_buffer.size )

                    if is_success
                        # Remove file
                        Rails.logger.info "Remove file: file: #{file.name} #{file.id}"
                        drive_service.delete_file(file.id)
                    end
                end
                page_token = response.next_page_token
            end while !page_token.nil?
        end

        def export(file_data, file_path, mime_type)
            Rails.logger.info "Exporting file '#{file_path}'.."
            drive = drive_service

            path = File.dirname(file_path)
            folder_id = create_folder(path, drive)

            metadata  = {
                name: File.basename(file_path),
                parents: [folder_id]
            }

            gd_file = drive.create_file(metadata, fields: 'id', upload_source: StringIO.new(file_data), content_type: mime_type)

            if gd_file.present?
                Rails.logger.info "Created file with id '#{gd_file.id}'"
            else
                Rails.logger.info "Unable to create file"
            end
        end

        private
        # Creates the given folder if its exists.
        # @return the folder id
        def create_folder( folder_name, drive_service )
            folder_id = nil

            # First check whether the folder exists.
            page_token = nil
            begin
                # A folder is a file with a special mimeType
                response = drive_service.list_files(q: "mimeType='application/vnd.google-apps.folder' and 'root' in parents",
                                                    spaces: 'drive',
                                                    fields:'nextPageToken, files(id, name)',
                                                    page_token: page_token)
                for file in response.files
                    # Process change
                    puts "Found file: #{file.name} #{file.id}"
                    next if file.name != folder_name

                    # Found file
                    folder_id = file.id
                end
                page_token = response.next_page_token
            end while !page_token.nil?

            # Create the folder if it not exists.
            if folder_id.nil?
                # Create folder
                folder_metadata = {
                    name: folder_name,
                    mime_type: 'application/vnd.google-apps.folder'
                }
                folder = drive_service.create_file(folder_metadata, fields: 'id')
                folder_id = folder.id
            end

            folder_id
        end

        def drive_service
            importer_data = @user.authenticated_importers.select { |importer_hash| importer_hash[:type] == GoogledriveImporter::IMPORTER_TYPE }
            auth_data = importer_data.first[:access_data]

            client_opts = JSON.parse(auth_data)
            auth_client = Signet::OAuth2::Client.new(client_opts)
            auth_client.fetch_access_token!
            drive = Google::Apis::DriveV3::DriveService.new
            drive.authorization = auth_client

            drive
        end
    end


end
