module BinectImporter
        class OneDriveImporter
                IMPORTER_TYPE = "onedrive"

                ACCESS_TOKEN    = "access_token"
                REFRESH_TOKEN   = "refresh_token"
                CLIENT_ID       = "client_id"
                REDIRECT_URI    = "redirect_uri"
                CLIENT_SECRET   = "client_secret"

                GRANT_TYPE_AUTHORIZATION = "authorization_code"

                ID_KEY                  = "id"
                DOWNLOAD_URL_KEY        = "@microsoft.graph.downloadUrl"
                NAME_KEY                = "name"
                SIZE_KEY                = "size"
                FILE_KEY                = "file"
                MIME_TYPE_KEY           = "mimeType"
                VALUE_KEY               = "value"
                FOLDER_KEY              = "folder"
                UPLOAD_URL_KEY          = "uploadUrl"
                ITEM_KEY                = "item"
                CONFLICT_BEHAVIOR_KEY   = "@microsoft.graph.conflictBehavior"

                CONFLICT_BEHAVIOR_RENAME = "rename"

                AUTHORIZATION           = "Authorization"
                AUTHORIZATION_PREFIX    = "Bearer"

                URI_ITEM_PREFIX                 = "https://graph.microsoft.com/v1.0/drive/items/"
                URI_TOKEN                       = "https://login.microsoftonline.com/common/oauth2/v2.0/token"
                URI_ROOT_CHILDREN               = "https://graph.microsoft.com/v1.0/drive/root/children"
                URI_CHILDREN_SUFFIX             = "children"
                URI_CREATE_UPLOAD_SUFFIX        = "createUploadSession"

                MIME_TYPE_JSON          = "application/json"
                MIME_TYPE_PDF           = "application/pdf"

                IMPORT_FOLDER_NAME      = "BinectImporter"

                CONTENT_LENGTH          = "Content-Length"
                CONTENT_RANGE           = "Content-Range"

                # Use 5 MiB as the fragment size.
                # Note that the fragment size must be a multitude of 320 KiB.
                FRAGMENT_SIZE           = 5 * 1024 * 1024

                UPLOAD_MAX_RETRIES      = 5
                UPLOAD_RETRY_WAIT       = 1

                UPLOAD_RESULT_FAILURE   = 0
                UPLOAD_RESULT_SUCCESS   = 1
                UPLOAD_RESULT_RETRY     = 2

                def initialize(user)
                        @user = user
                end

                def import
                        access_token = get_access_token(@user)

                        folder_id = create_folder(access_token, IMPORT_FOLDER_NAME)

                        if folder_id.nil?
                                Rails.logger.error "OneDriveImporter: Failed to retrieve the folder id."
                                return
                        end

                        folder_content_list = list_folder_contents(access_token, folder_id)

                        if folder_content_list.nil?
                                Rails.logger.error "OneDriveImporter: Failed to retrieve the folder contents."
                                return
                        end

                        folder_content_list.each do |file|
                                file_id         = file[OneDriveImporter::ID_KEY]
                                download_url    = file[OneDriveImporter::DOWNLOAD_URL_KEY]
                                file_name       = file[OneDriveImporter::NAME_KEY]
                                file_size       = file[OneDriveImporter::SIZE_KEY]
                                file_data       = file[OneDriveImporter::FILE_KEY]

                                mime_type = nil
                                if !file_data.nil?
                                        mime_type = file_data[OneDriveImporter::MIME_TYPE_KEY]
                                end

                                uri = URI(download_url)
                                file_content = Net::HTTP.get(uri)

                                Rails.logger.info "OneDriveImporter: Import PDF: ready"

                                is_success = yield( file_content, file_name, mime_type, file_size)

                                if is_success
                                        Rails.logger.info "OneDriveImporter: Remove file: file: #{file_name} #{file_id}"

                                        uri = URI("#{URI_ITEM_PREFIX}#{file_id}")

                                        request = Net::HTTP::Delete.new(uri)
                                        request[AUTHORIZATION] = "#{AUTHORIZATION_PREFIX} #{access_token}"

                                        http = Net::HTTP.new(uri.host, uri.port)
                                        http.use_ssl = true

                                        http.request(request)
                                end
                        end
                end

                def export(file_data, file_path, mime_type)
                        overall_retry_count = 0
                        while overall_retry_count < UPLOAD_MAX_RETRIES do
                                Rails.logger.info "OneDriveImporter: Exporting file '#{file_path}'"
                                path = File.dirname(file_path)

                                access_token = get_access_token(@user)

                                folder_id = create_folder(access_token, path)

                                if folder_id.nil?
                                        Rails.logger.error "OneDriveImporter: Failed to retrieve the folder id."
                                        return
                                end

                                # Start the upload session
                                file_name = File.basename(file_path)
                                uri = URI("#{URI_ITEM_PREFIX}#{folder_id}:/#{file_name}:/#{URI_CREATE_UPLOAD_SUFFIX}")

                                request = Net::HTTP::Post.new(uri)
                                request.content_type    = MIME_TYPE_JSON
                                request[AUTHORIZATION]  = "#{AUTHORIZATION_PREFIX} #{access_token}"

                                new_file_hash = {
                                        ITEM_KEY => {
                                                CONFLICT_BEHAVIOR_KEY => CONFLICT_BEHAVIOR_RENAME,
                                                NAME_KEY => file_name
                                        }
                                }

                                request.body = new_file_hash.to_json

                                http = Net::HTTP.new(uri.host, uri.port)
                                http.use_ssl = true

                                response_body = http.request(request).body
                                response_json = JSON.parse(response_body)

                                # Retrieve the upload URL for the session
                                upload_url = response_json[UPLOAD_URL_KEY]
                                upload_uri = URI(upload_url)

                                file_io_data = StringIO.new(file_data)

                                # Upload the file in chunks of the fragment size
                                upload_result = upload_file_data(file_io_data, upload_uri)

                                case upload_result
                                when UPLOAD_RESULT_SUCCESS
                                        Rails.logger.info "OneDriveImporter: Upload succeeded."
                                        return
                                when UPLOAD_RESULT_RETRY
                                        sleep UPLOAD_RETRY_WAIT

                                        Rails.logger.info "OneDriveImporter: Retry failed upload."
                                        overall_retry_count += 1
                                else
                                        Rails.logger.error  "OneDriveImporter: Upload failed."
                                        return
                                end
                        end

                        Rails.logger.error  "OneDriveImporter: Upload failed too many times."
                end

                private
                def get_access_token(user)
                        Rails.logger.info "OneDriveImporter: get_access_token"
                        return nil if user.nil?

                        Rails.logger.info "OneDriveImporter: get data.."
                        one_drive_data = user.authenticated_importers.select { |importer_hash| importer_hash[:type] == OneDriveImporter::IMPORTER_TYPE }

                        return nil if one_drive_data.nil? || one_drive_data.first.nil?

                        Rails.logger.info "OneDriveImporter: get new access token"
                        one_drive_data_hash = one_drive_data.first

                        refresh_token   = one_drive_data_hash[OneDriveImporter::REFRESH_TOKEN]
                        client_id       = one_drive_data_hash[OneDriveImporter::CLIENT_ID]
                        redirect_url    = one_drive_data_hash[OneDriveImporter::REDIRECT_URI]
                        client_secret   = one_drive_data_hash[OneDriveImporter::CLIENT_SECRET]

                        # Request the new access token.
                        begin
                                params = {
                                        OneDriveImporter::CLIENT_ID     => client_id,
                                        OneDriveImporter::REDIRECT_URI  => redirect_url,
                                        OneDriveImporter::CLIENT_SECRET => client_secret,
                                        OneDriveImporter::REFRESH_TOKEN => refresh_token,
                                        "grant_type"    => OneDriveImporter::REFRESH_TOKEN
                                }

                                uri = URI(URI_TOKEN)
                                result = Net::HTTP.post_form(uri, params)

                                response = JSON.parse(result.body)
                                access_token    = response[OneDriveImporter::ACCESS_TOKEN]
                                refresh_token   = response[OneDriveImporter::REFRESH_TOKEN]

                                importer_data_index = user.authenticated_importers.index { |importer_hash| importer_hash[:type] == OneDriveImporter::IMPORTER_TYPE }
                                importer_data = {
                                        class_name: "BinectImporter::OneDriveImporter",
                                        type: OneDriveImporter::IMPORTER_TYPE,
                                        OneDriveImporter::ACCESS_TOKEN => access_token,
                                        OneDriveImporter::REFRESH_TOKEN => refresh_token,
                                        OneDriveImporter::CLIENT_ID     => client_id,
                                        OneDriveImporter::REDIRECT_URI  => redirect_url,
                                        OneDriveImporter::CLIENT_SECRET => client_secret
                                }

                                if importer_data_index.nil?
                                        # Add new importer_data
                                        Rails.logger.info "Saving new OneDrive access token"
                                        user.authenticated_importers.push(importer_data)
                                else
                                        # Replace importer data
                                        Rails.logger.info "Updating OneDrive new access token"
                                        user.authenticated_importers[importer_data_index] = importer_data
                                end
                                unless user.save
                                        Rails.logger.error  "Unable to persist change: #{user.errors}"
                                        flash[:error] = "Unable to save data. Please try again"

                                        return nil
                                end

                                return access_token
                        rescue => e
                                Rails.logger.error "Failed to request the new access token: #{e}"
                                return nil
                        end
                end

                # Creates the given folder_name and returns its id if successful. Returns nil otherwise.
                def create_folder(access_token, folder_name)
                        if access_token.nil?
                                return nil
                        end

                        uri = URI(URI_ROOT_CHILDREN)
                        request = Net::HTTP::Post.new(uri)

                        request.content_type    = MIME_TYPE_JSON
                        request[AUTHORIZATION]  = "#{AUTHORIZATION_PREFIX} #{access_token}"

                        folder_hash = {
                                OneDriveImporter::NAME_KEY => folder_name,
                                OneDriveImporter::FOLDER_KEY => {}
                        }

                        request.body = folder_hash.to_json

                        http = Net::HTTP.new(uri.host, uri.port)
                        http.use_ssl = true

                        response_body = http.request(request).body
                        response_json = JSON.parse(response_body)

                        response_json[ID_KEY]
                end

                # Lists the contents of the given folder_id. Returns nil if the listing failed.
                def list_folder_contents(access_token, folder_id)
                        if access_token.nil? or folder_id.nil?
                                return nil
                        end

                        uri = URI("#{URI_ITEM_PREFIX}#{folder_id}/#{URI_CHILDREN_SUFFIX}")
                        request = Net::HTTP::Get.new(uri)

                        request.content_type    = MIME_TYPE_JSON
                        request[AUTHORIZATION]  = "#{AUTHORIZATION_PREFIX} #{access_token}"

                        http = Net::HTTP.new(uri.host, uri.port)
                        http.use_ssl = true

                        response_body = http.request(request).body
                        response_json = JSON.parse(response_body)

                        folder_list = response_json[VALUE_KEY]
                        folder_list.select do |item|
                                file = item[FILE_KEY]
                                !file.nil? && file[MIME_TYPE_KEY] == MIME_TYPE_PDF
                        end
                end

                # Uploads the given file_io_data in chunks for the given upload_uri.
                #
                # Returns the following result values;
                #
                #       - UPLOAD_RESULT_FAILURE = Upload failed.
                #       - UPLOAD_RESULT_SUCCESS = Upload finished successfully.
                #       - UPLOAD_RESULT_RETRY   = Upload needs to be retried from the beginning.
                def upload_file_data(file_io_data, upload_uri)
                        # Cancel the upload session if necessary.
                        delete_request = Net::HTTP::Delete.new(upload_uri)
                        delete_http = Net::HTTP.new(upload_uri.host, upload_uri.port)
                        delete_http.use_ssl = true

                        file_size = file_io_data.size

                        upload_http = Net::HTTP.new(upload_uri.host, upload_uri.port)
                        upload_http.use_ssl = true
                        while not file_io_data.eof? do
                                position_start = file_io_data.pos
                                chunk = file_io_data.read FRAGMENT_SIZE
                                position_end = file_io_data.pos

                                if position_end < 1
                                        Rails.logger.error "OneDriveImporter: Failed to upload file fragment."

                                        # Cancel the upload session.
                                        delete_http.request(delete_request)
                                        return UPLOAD_RESULT_FAILURE
                                end

                                upload_request = Net::HTTP::Put.new(upload_uri)
                                upload_request[CONTENT_LENGTH]  = chunk.size
                                upload_request[CONTENT_RANGE]   = "bytes #{position_start}-#{position_end - 1}/#{file_size}"

                                upload_request.body = chunk

                                # Retry the chunk upload a couple of times in case of an error.
                                retry_count = 0
                                upload_success = false
                                while retry_count < UPLOAD_MAX_RETRIES do
                                        upload_response = upload_http.request(upload_request)

                                        response_code = upload_response.code
                                        if response_code[0] == "2"
                                                upload_response_body = upload_response.body
                                                Rails.logger.info "OneDriveImporter: Upload fragment response: #{upload_response_body}"

                                                # Move on to the next chunk.
                                                upload_success = true
                                                break
                                        elsif response_code == "404"
                                                # Start over from the beginning.
                                                return UPLOAD_RESULT_RETRY
                                        else
                                                # Retry.
                                                sleep UPLOAD_RETRY_WAIT

                                                Rails.logger.info "OneDriveImporter: Retry failed chunk upload."
                                                retry_count += 1
                                        end
                                end

                                if !upload_success
                                        # The chunk upload failed too many times.
                                        # Cancel the upload session and stop.
                                        delete_http.request(delete_request)
                                        return UPLOAD_RESULT_FAILURE
                                end
                        end

                        UPLOAD_RESULT_SUCCESS
                end
        end
end
